package com.sun.xml.internal.bind.v2.model.annotation;

import java.util.HashMap;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Modifier;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.Map;
import java.lang.annotation.Annotation;
import com.sun.xml.internal.bind.v2.runtime.Location;
import java.lang.reflect.InvocationHandler;

public class LocatableAnnotation implements InvocationHandler, Locatable, Location
{
    private final Annotation core;
    private final Locatable upstream;
    private static final Map<Class, Quick> quicks;
    
    public static <A extends Annotation> A create(A annotation, Locatable parentSourcePos) {
        if (annotation == null) {
            return null;
        }
        Class<? extends Annotation> type = annotation.annotationType();
        if (LocatableAnnotation.quicks.containsKey(type)) {
            return ((Quick)LocatableAnnotation.quicks.get(type)).newInstance(parentSourcePos, annotation);
        }
        ClassLoader cl = SecureLoader.getClassClassLoader(LocatableAnnotation.class);
        try {
            Class<Object> loadableT = Class.forName(type.getName(), false, cl);
            if (loadableT != type) {
                return annotation;
            }
            return (Annotation)Proxy.newProxyInstance(cl, new Class[] { type, Locatable.class }, new LocatableAnnotation(annotation, parentSourcePos));
        }
        catch (ClassNotFoundException e) {
            return annotation;
        }
        catch (IllegalArgumentException e2) {
            return annotation;
        }
    }
    
    LocatableAnnotation(Annotation core, Locatable upstream) {
        this.core = core;
        this.upstream = upstream;
    }
    
    @Override
    public Locatable getUpstream() {
        return this.upstream;
    }
    
    @Override
    public Location getLocation() {
        return this;
    }
    
    @Override
    public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
        try {
            if (method.getDeclaringClass() == Locatable.class) {
                return method.invoke(this, args);
            }
            if (Modifier.isStatic(method.getModifiers())) {
                throw new IllegalArgumentException();
            }
            return method.invoke(this.core, args);
        }
        catch (InvocationTargetException e) {
            if (e.getTargetException() != null) {
                throw e.getTargetException();
            }
            throw e;
        }
    }
    
    @Override
    public String toString() {
        return this.core.toString();
    }
    
    static {
        quicks = new HashMap();
        for (Quick q : Init.getAll()) {
            LocatableAnnotation.quicks.put(q.annotationType(), q);
        }
    }
}
